program UFLBP; { LBP Firmware updater}

{$IFDEF WINDOWS}
uses synaser,synautil,synaip,blcksock,dos,crt;
var
ser:TBlockSerial;
TheComPort : string;
IPAddr : string;
Socket : TUDPBlockSocket;
{$ELSE}
uses dos,crt;
var TheComPort : word;
{$ENDIF}

{$I SELECTC}
{$I SELECTIO}
{I SELECTP}
{$I SELECTPR}
{$I INTERFCE}

type configfile = file;

const
{block addresses are in wordaddress / 512}
StartBlock = 4;
StaticBufferSize = 512*4;  {512 instructions of 4 bytes each}

type
BBufType = array[0..StaticBufferSize-1] of byte;
BBufPtrType = ^BBufType;
RBufType = array[0..StaticBufferSize-1] of byte;
RBufPtrType = ^RBufType;

var
BasePort : word;
counter : integer;
EraseSize : word;
WriteSize  : word;
Count : word;
FileName : string;
CFile : configfile;
BBufPtr : BBufPtrType;
RBufPtr : RBufPtrType;
BBuf : BBufType;
RBuf : RBufType;
BBufIndex : integer;
RBufIndex : integer;
BChecksum, RChecksum : byte;
DoReadback : boolean;
NoCRC : boolean;
CardName : string;

procedure Cleanup;
begin
  if protocol = SSLBP then
  begin
    { in case crcs are disabled don't stopall (which changes the baud rates) until communitions are finished }
    if not SSLBPProgSync(axis) then BumOut('Sync error');
    SSLBPSetBaud(Axis,2500000);
    SSLBPSetNoCRC(Axis,0);
  end;
end;

procedure Error(err : integer);
begin
  writeln(errorrecord[err].errstr);
  Cleanup;
  halt(2);
end;

procedure FixCRT;
{ Fix re-direction when using CRT unit }
begin
  assign(input,'');
  reset(input);
  assign(output,'');
  rewrite(output);
end;

procedure WriteBlock(theaxis,add:word; ourbuffer : BBufPtrType);
var ourdouble : comp;
begin
  DoubleByteRec(ourdouble).Byte0 := (ourbuffer^[0]);
  DoubleByteRec(ourdouble).Byte1 := (ourbuffer^[1]);
  DoubleByteRec(ourdouble).Byte2 := (ourbuffer^[2]);
  DoubleByteRec(ourdouble).Byte3 := (ourbuffer^[3]);
  DoubleByteRec(ourdouble).Byte4 := (ourbuffer^[4]);
  DoubleByteRec(ourdouble).Byte5 := (ourbuffer^[5]);
  DoubleByteRec(ourdouble).Byte6 := (ourbuffer^[6]);
  DoubleByteRec(ourdouble).Byte7 := (ourbuffer^[7]);
  WriteDouble(add,ourdouble);
end;

procedure ReadBlock(theaxis,add:word; ourbuffer : RBufPtrType);
var data : comp;
begin
  data := ReadDouble(add);
  ourbuffer^[0] := CompByteRec(data).Byte0;
  ourbuffer^[1] := CompByteRec(data).Byte1;
  ourbuffer^[2] := CompByteRec(data).Byte2;
  ourbuffer^[3] := CompByteRec(data).Byte3;
  ourbuffer^[4] := CompByteRec(data).Byte4;
  ourbuffer^[5] := CompByteRec(data).Byte5;
  ourbuffer^[6] := CompByteRec(data).Byte6;
  ourbuffer^[7] := CompByteRec(data).Byte7;
end;

function BufferEmpty : boolean;
var
i : integer;
empty : boolean;
begin
  empty := true;
  for i := 0 to EraseSize{BufferSize}-1 do
  begin
    if BBuf[i] <> 0 then empty := false;
  end;
  BufferEmpty := empty;
end;

procedure OurWriteFlash(address : longint);
var
i,j : word;
begin
  FlashStart(Axis);
  for i := 0 to (EraseSize div WriteSize)-1 do {must wait after each row }
  begin
    SetOffset(Axis,address+(i*WriteSize));
    for j := 1 to (WriteSize div 8) do
    begin
      BBufPtr := @BBuf[BBufIndex];
      WriteBlock(Axis,BBufIndex mod WriteSize,BBufPtr);
      BBufIndex := BBufIndex+8;
    end;
     CommitWrite(Axis);
  end;
  FlashStop(Axis);
end;




procedure OurReadFlash(address : longint);
var
data : comp;
i,j : word;
begin
  FlashStart(Axis);
  for i := 0 to (EraseSize div WriteSize)-1 do
  begin
    SetOffset(Axis,address+(i*WriteSize));
    for j := 1 to (WriteSize div 8) do
    begin
      RBufPtr := @RBuf[RBufIndex];
      ReadBlock(Axis,RBufIndex mod WriteSize,RBufPtr);
      RBufIndex := RBufIndex+8;
    end;
  end;
  FlashStop(Axis);
end;

procedure OurEraseFlash(add : longint);
begin
  FlashStart(Axis);
  SetOffset(Axis,add);
  CommitErase(Axis);
  FlashStop(Axis)
end;

procedure BarfOut(s: string);
begin
  writeln;
  writeln(s{,chr(7)}); { chr(7) causes long delays on some systems }
  Cleanup;
  halt(2);
end;

function OpenForRead(var fp: configfile; name: string): boolean;
begin
  Assign(fp,Name);
  {$I-}
  Reset(fp);
  {$I+}
  OpenForRead := IOResult = 0;
end { Open };

procedure Usage;
begin
  writeln;
  writeln(' UFLBP - Writes ROM data to Generic');
  writeln;
  writeln(' USAGE - UFLBP file [r] [n] [SSLBP_Channel_#]');
  writeln(' Where file is a .bin file');
  writeln(' First four characters of file name must match target name');
  writeln(' optional r parameter does block readback for checking');
  writeln(' optional n parameter for no crc checking for SSLBP');
  writeln(' optional SSLBP Channel number');
  writeln;
  halt(2);
end;

function ReadString(var idx: word) : string;
var
sleng,count,index : word;
tstring : string;
begin
  index := idx;
  sleng := BBuf[index] * 256 + BBuf[index+1];
  index := index +2; {skip over string length}
  if sleng >255 then sleng := 255;
  tstring := '';
  for count := 1 to sleng do
  begin
    if BBuf[index] <> 0 then tstring := tstring + chr(BBuf[index]);
    index := index +1;
  end;
  idx := index;
  ReadString := tstring;
end;

procedure GetParms;
var
index : integer;
tempstr : string;
retcode : integer;
axisnumber : byte;
  begin
  DoReadBack := false;
  NoCRC := false;
  Axis := 0;
  if ParamCount < 1 then Usage;
  FileName := paramStr(1);
  if not OpenForRead(CFile,FileName) then
  begin
    writeln('Can''t open configuration file');
    halt(2);
  end;
  for index := 2 to 4 do
  begin
    if length(ParamStr(index)) > 0 then
    begin
      if upstring(ParamStr(index)) = 'R' then DoReadBack := true;
      if upstring(ParamStr(index)) = 'N' then NoCRC := true;
      val(ParamStr(index),axisnumber,retcode);
      if retcode = 0 then Axis := axisnumber;
    end;
  end;
end;

Procedure printRbuff;
var
index : word;
foostr : string;
begin
  for index := 1 to EraseSize{(BufferSize)} do
  begin
      Hexprint(RBuf[index],2);
    end;
  end;

Procedure ShowErrors;
var
index : word;
foostr : string;
begin
  for index := 0 to EraseSize{BufferSize}-1 do
  begin
    if BBuf[index] <> RBuf[index] then
    begin
      write('Data Error at byte ');
      HexPrint(index,2);
      write(': Expected ');
      HexPrint(BBuf[index],2);
      write(' but got back ');
      Hexprint(RBuf[index],2);
      writeln;
      writeln('Press return to continue');
      readln(foostr);
    end;
  end;
end;

procedure PRGDoIt;
var
bytesread,index : word;
bytecount : longint;
blocknum : integer;
retchar : char;
hstring : string;
begin
  writeln('Starting to program');
  reset(CFile,1);
  blocknum := 0;
  Count := 0;
  while not EOF(CFile) do
  begin
    bytesread := 0;
    BBufIndex := 0;
    RBufIndex := 0;
    for index := 0 to EraseSize-1 do BBuf[index] :=0;
    blockread(CFile,BBuf,EraseSize{BufferSize},bytesread);
    if (bytesread > 0) and (BufferEmpty = false) then
    begin
      if (blocknum >= StartBlock) then
      begin
        OurEraseFlash(blocknum*EraseSize);
        Write('EG# ',blocknum,'  ');
        OurWriteFlash(blocknum*EraseSize);
        Write('WG# ',blocknum,'  ');
        if DoReadback then
        begin
          Write('RG# ',blocknum,'  ');
          OurReadFlash(blocknum*EraseSize);
          ShowErrors;
        end;
        writeln;
      end;
    end
    else
    begin
      Writeln('Skipping Block# ',blocknum);
    end;
    { buffer not empty }
    blocknum := blocknum +1;
  end; { not eof }
  writeln;
end;

function VerifyFileName : boolean;
var location : byte;
P: PathStr;
D: DirStr;
N: NameStr;
E: ExtStr;
ok : boolean;
begin
  ok := false;
  P := Filename;
  FSplit(P,D,N,E);
  location := pos(upstring(CardName),upstring(N));
  if location = 1 then ok := true; { must start with card name }
  if upstring(E) <> '.BIN' then ok := false; { must have BIN in file name }
  VerifyFileName := ok;
end;

begin {main}
{  FixCrt;}
  GetParms;
  GetOurEnv;
  if TheBaudRate <> 115200 then
  begin
    writeln('Baud rate must be 115200');
    halt(2);
  end;
  CardName := 'Undetected Card';
  if NoCRC then Message := 'Setup Mode NoCRC' else Message := 'Setup Mode';
  if not InitializeInterface(message) then BarfOut(message);

  if Protocol = SSLBP then
  begin
    writeln('Protocol = SSLBP');
    if SSLBPMinorRevision < 34 then BarfOut('Needs SSLBP Revision 34 or greater');
    CardName := SSLBPReadCardName(Axis);
  end;
  if Protocol = LBP then
  begin
    writeln('Protocol = LBP');
    CardName := LBPReadCardName;
  end;
    writeln('Card Name ',CardName);
  if not(VerifyFileName) then barfout('Filename doesn''t match Card or is not a .bin file');
  FlashStart(Axis);
  WriteSize := GetWriteSize(Axis);
  EraseSize := GetEraseSize(Axis);
  writeln('WriteSize ',WriteSize,' EraseSize ',EraseSize);
  FlashStop(Axis);
  PRGDoIt;
  writeln('Programming complete!');
  Cleanup;
  halt(0);
end.
{ 6-12-2012 added NoCRC command line parameter and file name checking}
{ 5-1-2013 cleanup bug fixed }
{ 8-1-2013 fixed start block }
{ 8-29-2013 use staticbuffersize for arrays, replace buffersize with erasesize}
{ clear read buffer }